#--
#
# Author:
#   Sebastian Tello
#
# Copyright (c) 2003,2012 CORE Security Technologies, CORE SDI Inc.
# All rights reserved.
#
# This computer software is owned by Core SDI Inc. and is
# protected by U.S. copyright laws and other laws and by international
# treaties.  This computer software is furnished by CORE SDI Inc.
# pursuant to a written license agreement and may be used, copied,
# transmitted, and stored only in accordance with the terms of such
# license and with the inclusion of the above copyright notice.  This
# computer software or any other copies thereof may not be provided or
# otherwise made available to any other person.
#
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI Inc. BE LIABLE
# FOR ANY DIRECT,  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR
# CONSEQUENTIAL  DAMAGES RESULTING FROM THE USE OR MISUSE OF
# THIS SOFTWARE
#
#--
#
# An instance of this Agent class should be obtained from the EntityManager
# (lib.impact_manager.entity_manager.EntityManager)
# method: get_agent(name). An instance of this class is also an instance from
# Entity (lib.impact_manager.entities.entitity.Entity) so it inherits all its
# behavior
#
#--

from impact_manager.entities.entity import Entity
from impact_manager.entities.module import Module

class Agent(Entity):

    def is_installed(self):
        """Returns True if the Agent is currently installed False if it isn't"""
        ins = self.get_xml_property("is installed")
        if ins.lower() == 'false':
            return False
        return True

    def install_pcap(self, timeout=30):
        """ Installs the PCAP plugin on this agent. Returns True if after the
        process ends the plugin is installed. False otherwise. """
        mod = Module(self._workspace, "PCAP Plugin Install", self.get_name())
        self.run_module_sync(mod, timeout)
        return self.is_pcap_installed()

    def remove_pcap(self, timeout=30):
        """ Removes the PCAP plugin from this agent. Returns True if after the
        process ends the plugin is not installed. False otherwise """
        mod = Module(self._workspace, "PCAP Plugin Remove", self.get_name())
        self.run_module_sync(mod, timeout)
        return not self.is_pcap_installed()

    def is_pcap_installed(self):
        """ Returns True if the PCAP plugin is installed on this agent.
        False if it is not """
        pcap = self.get_xml_property("pcap")
        if pcap is None:
            return False
        return True

    def make_persistent(self, connection_method='Connect from target', port=0):
        """ Creates a new persistent agent. Returns the new agent entity.
        None if persistency failed. """

        mod = Module(self._workspace, "Make Agent Persistent", self.get_name())
        mod.add_parameter('CONNECTION METHOD', connection_method)
        mod.add_parameter('PORT', str(port))

        self.run_module_sync(mod, 300)
        commited = mod.get_commited_entities()
        for ent in commited:
            if ent.__class__.__name__ == 'Agent':
                return ent

    def disconnect(self, timeout=30):
        """ Attempts to disconnect the agent
        Equivalent to right-clicking on the agent and choosing 'Disconnect'
        when the agent is connected. """
        mod = Module(self._workspace, "Disconnect Agent")
        mod.add_parameter('TARGET', self.get_name())
        mod.sync_run(timeout)

    def connect(self, timeout=60):
        """ Attempts to connect the agent.
        Equivalent to right-clicking on the agent and choosing 'Connect'
        when the agent is disconnected. """
        mod = Module(self._workspace, "Connect Agent")
        mod.add_parameter('TARGET', self.get_name())
        mod.sync_run(timeout)

    def is_connected(self):
        """ Returns True if the Agent is connected, false if it is not """
        return self.get_property("connected") == "1"

    def uninstall(self, timeout=60):
        """ Asks localagent to uninstall this agent (an agent can't uninstall
        itself. Returns True if the agent is no longer installed,
        False otherwise """
        #an agent can uninstall itself, so this will be run by localagent
        mod = Module(self._workspace, "Uninstall Agent")
        mod.add_parameter("TARGET", self.get_name(), "entity_name:agent")
        mod.add_parameter("UNINSTALL ALL AGENTS", "false", "bool")
        mod.sync_run(timeout)
        return not self.is_installed()

    def is_persistent(self):
        """ Returns True if the agent is persistent, False if it is not. """
        pers = self.get_xml_property("persistency")
        if pers is not None and pers.lower() == 'true':
            return True
        return False

    def set_reconnection_policy(self, connection_type="Connect From",
                                retry_interval=30, retry_timeout=3600,
                                agent_port=0):
        """ Sets the reconnection policies to this agent. Equivalent to
        selecting 'Set Reconnection Policy' from the agent's context menu in
        Impact. Currently the connection_type possible values are 'None' and
        'Connect From' """

        mod = Module(self._workspace, "Set Agent Reconnection Policy")
        mod.add_parameter("TARGET", self.get_name())
        mod.add_parameter("Reconnection Policy/CONNECTION TYPE",
                          str(connection_type))
        mod.add_parameter("Reconnection Policy/RETRY INTERVAL",
                          str(int(retry_interval)))
        mod.add_parameter("Reconnection Policy/RETRY TIMEOUT",
                          str(int(retry_timeout)))
        mod.add_parameter("Reconnection Policy/AGENT_PORT",
                          str(int(agent_port)))
        mod.sync_run(60)

    def get_proxy_agent(self):
        """ Returns the name of the parent agent in the agents proxying chain"""
        return self.get_xml_property("proxy agent")

    def spawn_shell(self):
        mod = Module(self._workspace, "Shell", self.get_name())
        self.run_module(mod)
        return mod

    def spawn_mini_shell(self):
        mod = Module(self._workspace, "Mini Shell", self.get_name())
        self.run_module(mod)
        return mod

    def spawn_python_shell(self):
        mod = Module(self._workspace, "Python Shell", self.get_name())
        self.run_module(mod)
        return mod

    def get_file_browser(self):
        ws = self._workspace.get_com_handler()
        me = ws.getModuleExecution()
        fb = me.getFileBrowser(self.get_name(),0)
        return fb

    def run_module(self, mod):
        """ Asks this agent to run the given module. The parameter should be an
        instance of the lib.impact_manager.entities.module.Module class.
        The module is run asynchronously as a thread. """
        mod.set_agent(self.get_name())
        mod.set_workspace(self._workspace)
        mod.run()

    def run_module_sync(self, mod, timeout=30*60):
        """ Asks this agent to run the given module. The parameter should be an
        instance of the lib.impact_manager.entities.module.Module class.
        The module is run synchronously, so this method waits until the module's
        status is 'failed', 'aborted', 'finished', or 'stopped'. If the given
        timeout (in seconds) expires an exception is raised. If no timeout is
        provided it will be set to 30 minutes, you might deactivate the timeout
        by providing a negative value (not recommended) """
        mod.set_agent(self.get_name())
        mod.set_workspace(self._workspace)
        mod.sync_run(timeout)