#--
#
# Author:
#   Sebastian Tello
#
# Copyright (c) 2003,2012 CORE Security Technologies, CORE SDI Inc.
# All rights reserved.
#
# This computer software is owned by Core SDI Inc. and is
# protected by U.S. copyright laws and other laws and by international
# treaties.  This computer software is furnished by CORE SDI Inc.
# pursuant to a written license agreement and may be used, copied,
# transmitted, and stored only in accordance with the terms of such
# license and with the inclusion of the above copyright notice.  This
# computer software or any other copies thereof may not be provided or
# otherwise made available to any other person.
#
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI Inc. BE LIABLE
# FOR ANY DIRECT,  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR
# CONSEQUENTIAL  DAMAGES RESULTING FROM THE USE OR MISUSE OF
# THIS SOFTWARE
#
#--

from xml.dom import minidom

from impact_manager.impact_db_manager import ImpactDBManager

class Entity:

    #ENTITY_EVENTS SUBSCRIPTIONS
    EVENT_ENTITY_UPDATED = 2
    EVENT_ENTITY_DELETED = 3

    def __init__(self, workspace, entity_id):
        self._entity_id = entity_id
        self._workspace = workspace
        self._db = ImpactDBManager()

    def get_name(self):
        """ Returns entity's name """
        return self.__select("Name")

    def get_id(self):
        """ Returns the entity's id """
        return self._entity_id

    def get_class(self):
        """ Returns entity's class. E.g. 'agent' or 'host' """
        return self.__select("Class")

    def get_workspace(self):
        """ Returns the workspace entity used by this entity """
        return self._workspace

    def get_type(self):
        """ Returns entity's class. E.g. 'level3' if entity is agent """
        return self.__select("Type")

    def get_xml(self):
        """ Returns entity's XML descriptor """
        return self.__select("XML")

    def get_xml_property(self, prop_key):
        """ Gets a property from the Entity's XML. If the property is composed
        of more properties, returns a dictionary with the keys and values of
        the nested properties """
        xml = self.get_xml()
        try:
            doc = minidom.parseString(xml)
        except:
            doc = minidom.parseString(xml.encode("UTF-8"))
        root = doc.getElementsByTagName("entity")[0]
        for prop in root.getElementsByTagName("property"):
            if prop.hasAttribute("key") and prop.getAttribute("key") ==prop_key:
                return self.__get_data_from_property(prop)
        return None

    def get_xml_properties(self):
        """ Gets a property from the Entity's XML. If the property is composed
        of more properties, returns a dictionary with the keys and values of
        the nested properties """
        xml = self.get_xml()
        try:
            doc = minidom.parseString(xml)
        except:
            doc = minidom.parseString(xml.encode("UTF-8"))
        root = doc.getElementsByTagName("entity")[0]
        return self.__get_data_from_property(root)


    def get_property(self, prop_name):
        """ Returns the value of the entity given property. E.g. its
        display_alias. Returns None if the entity does not have the given
        property name """

        prop_name = self._db.addslashes(str(prop_name))

        query = "SELECT Value_Tx FROM TEntityProperty WHERE " + \
                "Workspace_Id = %d AND Entity_Id = %d AND Name = '%s';" \
                % (self._workspace.id, self._entity_id, prop_name)
        rs = self._db.execute(query)
        if rs.num_rows() == 0:
            return None
        return rs.value(0)

    def subscribe_to_entity_events(self, event_types, function,
                                   *additional_parameters):
        """Subscribe a function to events related to this entity.
        event_types: is a list of event types you want to be notified about
            e.g. [Entity.EVENT_ENTITY_UPDATED, Entity.EVENT_ENTITY_DELETED]
        function: the function that will be called when the event occurs
        additional_parameters: When the callback function is called, it will
            provide a dictionary parameter containing information about the
            event, you can give any other parameters you want to be passed
            besides the data dictionary

        The data dictionary passed back to the subscribed functions contains:
            'EVENT_TIME': float: number of seconds since the epoc
            'EVENT_TYPE': int: either Entity.EVENT_ENTITY_UPDATED,
                                      or Entity.EVENT_ENTITY_DELETED
            'WORKSPACE_ID': int: workspace_id,
            'TASK_ID': int: task_id,
            'ENTITY_ID': int: entity_id
            'ENTITY_NAME': string: name of the entity
            'ENTITY_CLASS': string: class of the entity

        Note: Keep in mind that you won't get event notifications related to
        other entities
        Example:
        def myfunc(greeting, data):
            print greeting
            print "your event data is:", data
        emgr = EntityManager(wsp)
        host = emgr.get_entity('/192.168.10.1')
        agent = emgr.get_agent('/agent(1)')
        ...
        host.subscribe_to_entity_events([Entity.EVENT_ENTITY_UPDATED],
                                        myfunc, "Hi There!")
        agent.subscribe_to_entity_events([Entity.EVENT_ENTITY_UPDATED,
                                          Entity.EVENT_ENTITY_DELETED],
                                          myfunc, "Hi There!")
        """
        entity_mgr_event_handler =self._workspace.get_entity_mgr_event_handler()
        entity_mgr_event_handler.subscribe_by_entity(self._entity_id,
                                                     event_types, function,
                                                     additional_parameters)

    def __select(self, field):
        query = "SELECT %s FROM TEntity WHERE Workspace_Id = %d AND Id = %d;" \
                % (field,self._workspace.id, self._entity_id)
        rs = self._db.execute(query)
        return rs.value(0)

    def __get_data_from_property(self, parent):
        if (len(parent.childNodes) == 1
                      and parent.firstChild.nodeType == minidom.Node.TEXT_NODE):
            return parent.firstChild.nodeValue
        elif len(parent.childNodes) == 0:
            return ""
        else:
            ret = {}
            for child in parent.childNodes:
                if child.nodeName == "property" and child.hasAttribute("key"):
                    ret[child.getAttribute("key")
                        ] = self.__get_data_from_property(child)
            return ret
        
    def __repr__(self):
        return "<Entity(%s):%s on Workspace:%d>" % (self.get_class(),
                                                    self.get_id(),
                                                    self._workspace.id)