#--
#
# Author:
#   Sebastian Tello
#
# Copyright (c) 2003,2012 CORE Security Technologies, CORE SDI Inc.
# All rights reserved.
#
# This computer software is owned by Core SDI Inc. and is
# protected by U.S. copyright laws and other laws and by international
# treaties.  This computer software is furnished by CORE SDI Inc.
# pursuant to a written license agreement and may be used, copied,
# transmitted, and stored only in accordance with the terms of such
# license and with the inclusion of the above copyright notice.  This
# computer software or any other copies thereof may not be provided or
# otherwise made available to any other person.
#
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI Inc. BE LIABLE
# FOR ANY DIRECT,  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR
# CONSEQUENTIAL  DAMAGES RESULTING FROM THE USE OR MISUSE OF
# THIS SOFTWARE
#
#--

import time
from xml.dom import minidom
from xml.dom.minidom import Document

from comtypes import COMError

from impact_manager.impact_lib_exceptions \
                             import ImpactModuleException, ImpactEntityException
from impact_manager.impact_db_manager import ImpactDBManager
from impact_manager.resources.module_execution_event_handler \
                                              import ModuleExecutionEventHandler

class Module:

    #ENTITY_EVENTS SUBSCRIPTIONS
    EVENT_ENTITY_CREATED = 1
    EVENT_ENTITY_UPDATED = 2
    EVENT_ENTITY_DELETED = 3

    def __init__(self, workspace, name, agent='localagent'):
        """ Requires a workspace entity, the module name, and the agent name"""
        self.__workspace = workspace
        self.__params = []
        self.__name = name
        self.set_agent(agent)
        self.__handler = ModuleExecutionEventHandler(self)
        self.__entity_events_subscription_buffer = []
        #The following list will keep track of entities commited by this module
        self.__commited_entities = []
        self.subscribe_to_entity_events([Module.EVENT_ENTITY_CREATED],
                                        self.__on_entity_commited)

    def add_parameter(self, key, value, ptype=None):
        """ Adds a parameter to the xml param file. If ptype is not given, the
        parameter type will be obtained from the module's xmldata, if it is not
        found, then 'string' will be used. Possible ptypes include:
        string, uint16, bool, entity_name:[host], user:[option 1,option 2..],
        unicode, file, parameters (in this case, value should be a list of
        dictionaries of keys, values and types).
        The easiest way to add complex (advanced) parameters is to provide the
        key path like in this example:
        mod.add_parameter("AGENT CONNECTION/CONNECTION_METHOD",
                          "Reuse connection")
        """
        self.__params.append({'key':key, 'value':value, 'type':ptype})

    def get_task_id(self):
        """ Returns the Impact's taks id of this module execution instance"""
        return self.__handler.get_task_id()

    def set_task_id(self,task_id):
        """ Sets 'by hand' the task id for this module """
        self.__handler.set_task_id(int(task_id))

    def get_execution_parameters(self):
        """ Returns a dictionary with param_name:param_value pairs. Representing
        the parameters used to run the module"""
        db = ImpactDBManager()
        query = "SELECT Name, Value_Tx " \
                "FROM TParameter " \
                "WHERE Workspace_Id = %s " \
                "AND Task_Id = %d;" % (self.__workspace.id, self.get_task_id())

        rs = db.execute(query)
        params = {}
        while not rs.EOF:
            params[rs.value('Name')] = rs.value('Value_Tx')
            rs.next()
        return params

    def get_child_modules(self):
        """Returns a list of child modules"""
        db = ImpactDBManager()
        query = """ SELECT A.Task_Id
                    FROM TTask as A, TTaskRelation as B
                    WHERE A.Workspace_Id = %s
                    AND B.ParentTask_Id = %s
                    AND A.Task_Id = B.Task_Id
                    AND A.Workspace_Id = B.Workspace_Id;""" % (
                                        self.__workspace.id, self.get_task_id())
        rs = db.execute(query)
        ret = []
        while not rs.EOF:
            ret.append(self.__workspace.task(rs.value(0)))
            rs.next()
        return ret

    def get_commited_entities(self):
        """Returns an array with Entity and/or Agent instances representing the
        entities commited by this module"""
        return self.__commited_entities

    def get_commited_agents(self):
        """Returns an array with Agent instances representing the agent entities
        commited by this module"""
        ret = []
        for entity in self.__commited_entities:
            if entity.get_class() == 'agent':
                ret.append(entity)
        return ret

    def get_qty_commited_agents(self):
        """Returns qty of agents commited by the module """
        return len(self.get_commited_agents())

    def subscribe_to_entity_events(self, event_types, function,
                                                        *additional_parameters):
        """Subscribe a function to entity events triggered by this module run.
        (EVENTS GENERATED BY THIS TASK OR ANY OF ITS DESCENDANTS)

        event_types: is a list of event types you want to be notified about
            e.g. [Module.EVENT_ENTITY_CREATED, Module.EVENT_ENTITY_DELETED]
        function: the function that will be called when the event occurs
        additional_parameters: When the callback function is called, it will
            provide a dictionary parameter containing information about the
             event, you can give any other parameters you want to be passed
             besides the data dictionary

        The data dictionary passed back to the subscribed functions contains:
            'EVENT_TIME': float: number of seconds since the epoc
            'EVENT_TYPE': int: either Module.EVENT_ENTITY_CREATED,
                                      Module.EVENT_ENTITY_UPDATED,
                                      or Module.EVENT_ENTITY_DELETED
            'WORKSPACE_ID': int: workspace_id,
            'TASK_ID': int: task_id,
            'ENTITY_ID': int: entity_id
            'ENTITY_NAME': string: name of the entity
            'ENTITY_CLASS': string: class of the entity

        Note: Keep in mind that you won't get event notifications generated by
        other impact tasks. Example:

        def myfunc(greeting, data):
            print greeting
            print "your event data is:", data
        mod = Module(ws, "New Host")
        ...
        mod.subscribe_to_entity_events([Module.EVENT_ENTITY_CREATED],
                                        myfunc,
                                        "Hi There!")
        mod.run()
        """
        if self.get_task_id() is None:
            #we cannot subscribe right now, as we don't have a task id yet
            #but we buffer this call so we subscribe as soon as we get one
            #(the module_execution_event_handler will let us know when)
            self.__entity_events_subscription_buffer.append((event_types,
                                               function, additional_parameters))
        else:
            #we already have a task id, then proceed with the subscription
            entity_mgr_event_handler = self.__workspace \
                                                 .get_entity_mgr_event_handler()
            entity_mgr_event_handler.subscribe_by_task_or_descendants(
               self.get_task_id(), event_types, function, additional_parameters)

    def commit_buffered_entity_events_subscriptions(self):
        """This method is used internally by the lib. Don't invoke it"""
        if self.get_task_id() is not None:
            entity_mgr_event_handler = self.__workspace \
                                                 .get_entity_mgr_event_handler()
            while len(self.__entity_events_subscription_buffer) > 0:
                sub = self.__entity_events_subscription_buffer.pop(0)
                entity_mgr_event_handler.subscribe_by_task_or_descendants(
                                     self.get_task_id(), sub[0], sub[1], sub[2])

    def get_name(self):
        """ Returns the module name """
        return self.__name

    def get_category(self):
        """ Returns the module's category """
        name = self.get_name()

        db = ImpactDBManager()
        name = db.addslashes(name)

        query = "SELECT Category " \
                "FROM TModule " \
                "WHERE ModuleName = '%s';" % (name)
        rs = db.execute(query)
        return rs.value(0)

    def get_xml_data(self):
        """ Returns the module's __xmldata__ """
        name = self.get_name()

        db = ImpactDBManager()
        name = db.addslashes(name)

        query = "SELECT XML " \
                "FROM TModule " \
                "WHERE ModuleName = '%s';" % (name)
        rs = db.execute(query)
        return rs.value(0)

    def get_xml_property_value(self, property_key):
        """ Returns an array of strings with the values found in __xmldata__ for
        the given property key. An empty array is returned if the property
        wasn't found or it has no data.
        E.g.: my_module.get_xml_property_value('supported systems')"""
        xml = self.get_xml_data()
        doc = minidom.parseString(xml.encode( "utf-8" ) )
        ret = []
        properties = doc.getElementsByTagName("property")
        for prop in properties:
            if (prop.hasAttribute("key") and
                      prop.getAttribute("key").lower() == property_key.lower()):

                ret.extend(Module.__get_xml_data(prop))
        return ret


    def get_xml_data_as_dict(self, from_prop_key=None):
        """ Gets all property from the module's XML. If the property is composed
        of more properties, returns a dictionary with the keys and values of
        the nested properties """
        xml = self.get_xml_data()
        try:
            doc = minidom.parseString(xml)
        except:
            doc = minidom.parseString(xml.encode("UTF-8"))
        root = doc.getElementsByTagName("entity")[0]

        if from_prop_key:
            for prop in root.getElementsByTagName("property"):
                if (prop.hasAttribute("key")
                and prop.getAttribute("key") == from_prop_key):
                    return Module.__get_data_from_property(prop)
            return None
        return Module.__get_data_from_property(root)


    def get_module_path(self):
        """ Returns the module's full path """
        name = self.get_name()
        db = ImpactDBManager()
        name = db.addslashes(name)

        query = "SELECT f.Path FROM TModule AS m "\
                "JOIN TModuleFile AS f  ON m.FileId = f.FileId " \
                "WHERE m.ModuleName = '%s'" % name

        rs = db.execute(query)
        return rs.value(0)

    def run(self):
        """ Executes the module """
        try:
            wrapper = self.__handler.wrap_me()
            me = self.__workspace.get_module_execution_com()
            me.runModule(self.__agent, self.__name, self.get_params_xml(),
                                                                        wrapper)

        except COMError, e:
            self.__handler.set_status_failed()
            raise ImpactModuleException(e.details[0])
        except Exception, e:
            self.__handler.set_status_failed()
            raise ImpactModuleException("Unexpected Exception: %s" % str(e))



    def sync_run(self, timeout=60*30):
        """ Executes the module and waits until it finishes or it times out"""
        start_time = time.time()
        self.run()
        while self.get_status() not in ['finished',
                                        'stopped',
                                        'aborted',
                                        'failed']:
            elapsed_time = time.time() - start_time
            if timeout >= 0 and elapsed_time > timeout:
                raise ImpactModuleException("Timeout expired (%d secs)"
                                                                      % timeout)
            time.sleep(5)
        return self.get_status() == 'finished'


    def stop(self):
        """ Attemps to stop the module """
        if not (self.__handler is None):
            try:
                me = self.__workspace.get_module_execution_com()
                me.stopModule(self.get_task_id())
            except:
                pass

    #TODO: Refactor all this and the impact event handler in order to have a
    #single general event handler for all modules.

    def resume(self, params={}, use_handler=True):
        """ Attempts to resume the module. Returns a Module instance
        of the new cloned module and an Module instance of the module
        that resumed it """

        sta = self.get_status()
        if sta not in ['finished', 'aborted', 'stopped']:
            raise ImpactModuleException("Can't resume a task in '%s' state"
                                                                          % sta)
        emgr = self.__workspace.entity_manager
        res_task_name = "/" + str(self.get_task_id())
        try:
            res_task = emgr.get_entity(res_task_name)
        except ImpactEntityException:
            raise ImpactModuleException("Task is not resumable.")

        #PREPARE A MODULE INSTANCE OF THE MODULE TO BE RE RUN
        resumed_module = Module(self.__workspace, self.__name, self.__agent)
        if use_handler:
            handler = ModuleExecutionEventHandler(resumed_module, True)
            resumed_module.__handler = handler
            me = self.__workspace.get_module_execution_com()
            me.registerGeneralHandler(handler.wrap_me())

        #PREPARE THE TASK RESUMER
        mod_name = res_task.get_xml_property("runner")
        resumer = Module(self.__workspace, mod_name)
        resumer.add_parameter("TARGET", res_task_name)
        for k,v in params.items():
            resumer.add_parameter(k, v)

        #LAUNCH RESUMER
        resumer.sync_run()

        if not use_handler:
            tasks = self.__workspace.get_tasks_list()
            tasks.reverse()
            for t in tasks:
                if t[0] != self.get_task_id() and t[1] == self.get_name():
                    resumed_module = self.__workspace.task(t[0])
                    break

        #RETURN THE RESUMED MODULE
        return resumed_module, resumer

    def get_status(self):
        """ Possible status to return:
        not started: the modules has't been run yet
        failed: there was an attempt to run the module, but something went wrong
        initial: finished: stopped: aborted: stopping: running: refers to the
                                                     corresponding impact status
        """
        return self.__handler.get_status()

    def get_status_from_db(self):
        """ It returns the status but from a query in the DB
        Possible status to return:
        6 = finished
        5 = aborted
        4 = running
        3 = stopped
        """
        db = ImpactDBManager()
        query = """ SELECT "Status"
                    FROM "dbo"."TTask"
                    WHERE Workspace_Id = %s
                    AND Task_Id = %s
                """ % (self.__workspace.id, self.get_task_id())
        rs = db.execute(query)

        if rs.value('Status') == 3:
            return "stopped"
        if rs.value('Status') == 4:
            return "running"
        if rs.value('Status') == 5:
            return "aborted"
        if rs.value('Status') == 6:
            return "finished"


    def get_start_time(self):
        """ Returns the module's start time as stored in the database.
        Returns None if the modules hasn't been run yet. """
        task_id = self.get_task_id()
        ws_id = self.__workspace.id

        if task_id is None:
            return None

        db = ImpactDBManager()
        query = "SELECT Started " \
                "FROM TTask " \
                "WHERE Workspace_Id = %d " \
                "AND Task_Id = %d;" % (ws_id, task_id)
        rs = db.execute(query)
        return rs.value(0)

    def get_end_time(self):
        """ Returns the module's finish time as stored in the database.
        Returns None if the modules hasn't been run yet. """
        task_id = self.get_task_id()
        ws_id = self.__workspace.id

        if task_id is None:
            return None

        db = ImpactDBManager()
        query = "SELECT Finished " \
                "FROM TTask " \
                "WHERE Workspace_Id = %d " \
                "AND Task_Id = %d;" % (ws_id, task_id)
        rs = db.execute(query)
        return rs.value(0)

    def get_log(self, tail_lines=0, detail='high'):
        """Returns the last tail_lines lines from the log in the detail level
        provided. If tail_lines is not provided (or equal or lower than zero)
        the whole log is returned.
        Possible detail levels are: 'high', 'medium, 'low'.
        If it is not provided 'high' is used.
        If the module hasn't been started yet, None is returned """

        log_level = 2 #high is default
        if detail.lower() == 'medium':
            log_level = 1
        elif detail.lower() == 'low':
            log_level = 0

        task_id = self.get_task_id()
        ws_id = self.__workspace.id

        if task_id is None:
            return None

        db = ImpactDBManager()

        limit = ""
        if tail_lines > 0:
            limit = "TOP %d" % (int(tail_lines))

        query = "SELECT %s Text_Tx " \
                "FROM TLogLine " \
                "WHERE Workspace_Id = %d " \
                "AND Task_Id = %d " \
                "AND TraceLevel <= %d " \
                "ORDER BY Line_Nr DESC" % (limit , ws_id, task_id, log_level)

        rs = db.execute(query)
        ret = ""
        while not rs.EOF:
            line = rs.value(0)
            if line[-1] == '\x01':
                #line ending with \01. Should be concatenated with the following
                ret = "%s%s" % (line[:-1], ret)
            else:
                ret = "%s\n%s" % (line, ret)
            rs.next()
        return ret

    def get_xml_output(self):
        """ Returns the module's XML output as stored in the database.
        Returns None if the modules hasn't been run yet. """
        task_id = self.get_task_id()
        ws_id = self.__workspace.id

        if task_id is None:
            return None

        db = ImpactDBManager()

        query = "SELECT \"Output_XML\"" \
                "FROM \"dbo\".\"TOutput\"" \
                "WHERE Workspace_Id = %d and Task_Id = %d" % (ws_id, task_id)

        rs = db.execute(query)
        ret = rs.value(0)
        if ret is None: ret = ""
        return ret

    def get_params_xml(self, include_default_params=True):
        """ Returns all the parameters definition (which were set with
        add_parameter(..)) as XML """

        doc = None
        if not include_default_params:
            xml = ""
        else:
            #get parameters xml with default values
            xml = self.__get_default_params_xml()

        #try to parse that. If it is not possible create a new xml from scratch
        try:
            doc = minidom.parseString(xml)
        except:
            doc = Document()
            root = doc.createElement("property")
            root.setAttribute("type", "parameters")
            root.setAttribute("key", "parameters")
            doc.appendChild(root)
        root = doc.firstChild

        self.__attach_properties(doc, root, self.__params)
        return root.toxml()

    def get_workspace(self):
        """ Returns the workspace entity set to this module """
        return self.__workspace

    def set_workspace(self, workspace):
        """ Changes the current module's workspace.
        The parameter should be an impact_manager.Workspace instance """
        self.__workspace = workspace

    def set_agent(self, agent_name):
        """Changes the current module's associated agent. Agent name is a str"""
        if agent_name == "localagent":
            self.__agent = agent_name
            return
        emg = self.__workspace.entity_manager
        agents = emg.list_agents()
        for agent in agents:
            if str(agent_name) in agent:
                self.__agent = agent
                return
        self.__agent = agent_name

    def get_agent(self):
        """Returns the agent name associated to this module"""
        return self.__agent

    def __get_default_params_xml(self):
        """Returns the section of the xmldata containing the parameters and
        their default values """
        xmldata = self.get_xml_data()
        if xmldata is None:
            return ""
        doc = minidom.parseString(xmldata)
        parameters = None
        properties = doc.getElementsByTagName("property")
        for prop in properties:
            if (prop.hasAttribute("key")
            and prop.getAttribute("key") == "parameters"):
                parameters = prop
                break
        if parameters is not None:
            return parameters.toxml()
        return ""

    def __on_entity_commited(self, data):
        from impact_manager.entities.entity import Entity
        from impact_manager.entities.agent import Agent
        if data['ENTITY_CLASS'] == 'agent':
            en = Agent(self.__workspace, data['ENTITY_ID'])
        else:
            en = Entity(self.__workspace, data['ENTITY_ID'])
        self.__commited_entities.append(en)


    @staticmethod
    def __get_xml_data(node):
        ret = []
        if node.nodeType == minidom.Node.TEXT_NODE:
            ret.append(node.data)
        for child in node.childNodes:
            ret.extend(Module.__get_xml_data(child))
        return ret


    def __attach_properties(self, document, parent_node, properties):
        advanced_properties = []
        for p in properties:
            if '/' in str(p['key']):
                ap = {'key':p['key'], 'value':p['value'], 'type':p['type']}
                advanced_properties.append(ap)
                continue
            try:
                node = Module.__search_property(parent_node, str(p['key']))
                if not node:
                    node = document.createElement("property")
                    #string is the default type
                    _type = "string"
                    if p['type']:
                        _type = str(p['type'])
                    node.setAttribute("type", _type)
                    node.setAttribute("key", str(p['key']))
                else:
                    #overwrite type attribute if given
                    if p.get('type'):
                        node.setAttribute("type", str(p['type']))

                tp = node.getAttribute("type")
                #This hack, transforms an entity display name to the entity name
                #as expected by the module
                if tp.startswith("entity_name:") and p['value']:
                    p['value'] = self.__get_entity_name(tp, p['value'])

                if tp.lower() == "parameters":
                    self.__attach_properties(document, node, p['value'])

                #parameters with type OS behave funny. The value of it
                #is not stored as a Text Node, but in an attribute 'name'
                #Besides the xml_data in the database (for New Host)
                #defers from the real xml_data in the module file.
                #For this parameter to work we need to remove its child nodes
                elif tp.lower() == "os":
                    node.setAttribute("name", str(p['value']))
                    for child in node.childNodes[:]:
                        node.removeChild(child)
                else:
                    child = document.createTextNode(str(p['value']))
                    if node.hasChildNodes():
                        node.replaceChild(child, node.firstChild)
                    else:
                        node.appendChild(child)
            except Exception:
                pass
            parent_node.appendChild(node)

        for p in advanced_properties:
            keys = p['key'].split('/')
            node = parent_node
            for k in keys:
                node = Module.__search_property(node, k)
                if not node:
                    break
            if node:
                tp = node.getAttribute("type")

                #This hack, transforms an entity display name to the entity name
                #as expected by the module
                if tp.startswith("entity_name:") and p['value']:
                    p['value'] = self.__get_entity_name(tp, p['value'])
                #parameters with type OS behave funny. The value of it
                #is not stored as a Text Node, but in an attribute 'name'
                #Besides the xml_data in the database (for New Host)
                #defers from the real xml_data in the module file.
                #For this parameter to work we need to remove its child nodes
                if tp.lower()== "os":
                    node.setAttribute("name", str(p['value']))
                    for child in node.childNodes[:]:
                        node.removeChild(child)
                else:
                    child = document.createTextNode(str(p['value']))
                    if node.hasChildNodes():
                        node.replaceChild(child, node.firstChild)
                    else:
                        node.appendChild(child)
                if p['type']:
                    node.setAttribute("type", str(p['type']))
            else:
                #add the node structure manually
                cur_node = parent_node
                last = keys[-1]
                for k in keys[:-1]:
                    node = Module.__search_property(cur_node, k)
                    if not node:
                        node = document.createElement("property")
                        node.setAttribute("key", k)
                        node.setAttribute("type", "parameters")
                        cur_node.appendChild(node)
                    cur_node = node
                child = document.createElement("property")
                child.setAttribute("key", last)
                _type = "string"
                if p['type']:
                    _type = str(p['type'])
                child.setAttribute("type", _type)
                text = document.createTextNode(str(p['value']))
                child.appendChild(text)
                cur_node.appendChild(child)

    def __get_entity_name(self, ptype, value):
        entity_classes = ptype[len("entity_name:"):].split(',')
        entity_classes = [c.strip() for c in entity_classes]
        emgr = self.__workspace.entity_manager
        #first check if the given value is already an entity_name and its
        #class in the class list of the parameter type
        try:
            ent = emgr.get_entity(value)
            if not entity_classes:
                return value
            if ent.get_class() in entity_classes:
                return value
        except ImpactEntityException:
            pass

        #second check if the given value is a display name
        try:
            en = emgr.get_entity_name(value, entity_classes)
            if en:
                return en
        except Exception:
            pass

        #if not, return the same value sent
        return value


    @staticmethod
    def __search_property(parent_node, key):
        for child in parent_node.childNodes:
            if (child.nodeType == minidom.Node.ELEMENT_NODE
            and child.tagName == "property"
            and child.hasAttribute("key")
            and child.getAttribute("key") == key):
                return child
        return None

    @staticmethod
    def __get_data_from_property(parent):
        ret = {}
        for k,v in parent.attributes.items():
            ret[k] = v
        ret["children"] = []
        ret["value"] = ""
        for child in parent.childNodes:
            if child.nodeType == minidom.Node.TEXT_NODE:
                ret["value"] = child.nodeValue
            elif child.nodeType == minidom.Node.ELEMENT_NODE:

                ret["children"].append(
                                    Module.__get_data_from_property(child))
        return ret
    
    def __repr__(self):
        w = ""
        if self.__workspace:
            w = " on Workspace:%s" % self.__workspace.id
        if self.get_task_id():
            return "<Module:'%s':Task:%s%s>" % (self.__name,self.get_task_id(),w)
        else:
            return "<Module:'%s'%s>" % (self.__name, w)
