#--
#
# Author:
#   Sebastian Tello
#
# Copyright (c) 2003,2012 CORE Security Technologies, CORE SDI Inc.
# All rights reserved.
#
# This computer software is owned by Core SDI Inc. and is
# protected by U.S. copyright laws and other laws and by international
# treaties.  This computer software is furnished by CORE SDI Inc.
# pursuant to a written license agreement and may be used, copied,
# transmitted, and stored only in accordance with the terms of such
# license and with the inclusion of the above copyright notice.  This
# computer software or any other copies thereof may not be provided or
# otherwise made available to any other person.
#
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI Inc. BE LIABLE
# FOR ANY DIRECT,  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR
# CONSEQUENTIAL  DAMAGES RESULTING FROM THE USE OR MISUSE OF
# THIS SOFTWARE
#
#--

import time

from impact_manager.workspace_manager import WorkspaceManager
from impact_manager.impact_com_manager import ImpactComManager
from impact_manager.impact_db_manager import ImpactDBManager
from impact_manager.impact_lib_exceptions import ImpactSchedulerException, \
                                                 ImpactModuleException
from impact_manager.entities.module import Module

class ScheduledTask:

    def __init__(self, name, time_str, frequency, description):
        """ initializes a Schedule instance
        name: task name
        time_str: string with the format yyyy/mm/dd hh:mm
        frequency: string [once|day|working day|mon wed fri|tue thu|week|month]
        description: string with any description """
        self.__sched_time = time.strptime(time_str, "%Y/%m/%d %H:%M")
        freqs = {'once':'once',
                 'day':'daily',
                 'working day':'working',
                 'mon wed fri':'monwedfri',
                 'tue thu':'tuethu',
                 'week':'weekly',
                 'month':'monthly'}
        self.__name = name
        self.__frequency = freqs[frequency.lower()]
        self.__description = description
        self.__mod_name = None
        self.__mod_params = {}
        self.__task_id = None

    def __second_init__(self, task_id):
        self.__task_id = int(task_id)

        db = ImpactDBManager()
        rs = db.execute("SELECT * FROM TScheduledTasks WHERE Id = %d;" %task_id)

        if rs.num_rows() == 0:
            raise ImpactSchedulerException(
                          "Could not find existing scheduled task with id '%d'."
                          % task_id)

        st_info = rs.get_row_as_dict()

        self.__description = st_info['Description']
        self.__mod_name = st_info['Module_Name']
        self.__name = st_info['Name']

    @staticmethod
    def get_scheduled_tasks():
        db = ImpactDBManager()
        rs = db.execute("SELECT Id FROM TScheduledTasks;")
        ret = []
        while not rs.EOF:
            st = ScheduledTask('', '2010/02/14 00:00', 'once', '')
            st.__second_init__(rs.value('Id'))
            ret.append(st)
            rs.next()
        return ret

    @staticmethod
    def get_all_executed_scheduled_tasks():
        """Returns a list o all the executed scheduled tasks from impact"""
        est = []
        sts = ScheduledTask.get_scheduled_tasks()
        for st in sts:
            est.extend(st.get_executed_scheduled_tasks())
        return est

    def get_name(self):
        return self.__name

    def get_description(self):
        return self.__description

    def get_module_name(self):
        return self.__mod_name

    def start_now(self):
        """Attempts to initiate a new execution of this task"""
        if self.__task_id is None:
            raise ImpactSchedulerException("There's no scheduled task to start")
        try:
            impact_com = ImpactComManager().get_impact_com()
            scheduler = impact_com.getScheduler()
            scheduler.runScheduledTask(self.__task_id)
        except Exception, e:
            raise ImpactSchedulerException("Error starting scheduled task %d: %s"
                                           % (self.__task_id, str(e)))
        except:
            raise ImpactSchedulerException(
                                   "Unexpected error starting scheduled task %d"
                                   % self.__task_id)

    def delete(self):
        """Attempsts to delete this task (if a schedule_x method has been
        called previously) if no schedule_x method has been called or something
        goes wrong throws an ImpactModuleException."""
        if self.__task_id is None:
            raise ImpactSchedulerException(
                                         "There is no scheduled task to delete")
        try:
            impact_com = ImpactComManager().get_impact_com()
            scheduler = impact_com.getScheduler()
            scheduler.deleteScheduledTask(self.__task_id)
            self.__task_id = None
        except Exception, e:
            raise ImpactSchedulerException("Error deleting scheduled task %d: %s"
                                           % (self.__task_id, str(e)))
        except:
            raise ImpactSchedulerException(
                                   "Unexpected error deleting scheduled task %d"
                                   % self.__task_id)

    def get_executed_scheduled_tasks(self):
        """Returns a list of ExecutedScheduledTask entities associated with this
        scheduled task"""
        if self.__task_id is None:
            raise ImpactSchedulerException(
                                  "The scheduled task has not been created yet")
        est_list = []

        db = ImpactDBManager()
        rs = db.execute("get_executed_scheduled_tasks")

        while not rs.EOF:
            if int(self.__task_id) == rs.value("Scheduled_Task_Id"):
                est = ExecutedScheduledTask(rs.value("Workspace_Id"),
                                            self.__name)
                est_list.append(est)
            rs.next()
        return est_list


    def schedule_network_vulnerability_test(self, target_range):
        if self.__task_id is not None:
            raise ImpactSchedulerException("This task already has a schedule. "
                                           "Create a new scheduled task.")
        self.__mod_name = "Network Vulnerability Test"
        self.__mod_params['TARGETRANGE'] = str(target_range)
        self.__commit_schedule()

    def schedule_client_side_vulnerability_test(self, email_file,
                                   wait_client_connection_hs, application_name):
        if self.__task_id is not None:
            raise ImpactSchedulerException("This task already has a schedule. "
                                           "Create a new scheduled task.")
        self.__mod_name = "Client-side Vulnerability Test"
        self.__mod_params['APPLICATION'] = str(application_name)
        self.__mod_params['EMAILFILE'] = str(email_file)
        self.__mod_params['DURATIONTIME'] = str(wait_client_connection_hs)
        self.__commit_schedule()

    def schedule_vsv_languard(self, file_to_import):
        if self.__task_id is not None:
            raise ImpactSchedulerException("This task already has a schedule. "
                                           "Create a new scheduled task.")
        self.__schedule_vsv_only_file(file_to_import, "languard")
        self.__commit_schedule()

    def schedule_vsv_ibm_internet_scanner(self, dsn, dbpass, dbuser, job_id):
        if self.__task_id is not None:
            raise ImpactSchedulerException("This task already has a schedule. "
                                           "Create a new scheduled task.")
        self.__mod_name = "Vulnerability Scanner Validator"
        self.__mod_params['IMPORTER'] = "iss"
        self.__mod_params['DBPASS'] = str(dbpass)
        self.__mod_params['DBUSER'] = str(dbuser)
        self.__mod_params['DSN'] = str(dsn)
        self.__mod_params['JOBID'] = int(job_id)
        self.__commit_schedule()

    def schedule_vsv_ncircle(self, file_to_import):
        if self.__task_id is not None:
            raise ImpactSchedulerException("This task already has a schedule. "
                                           "Create a new scheduled task.")
        self.__schedule_vsv_only_file(file_to_import, "ncircle")
        self.__commit_schedule()

    def schedule_vsv_nessus(self, file_to_import):
        if self.__task_id is not None:
            raise ImpactSchedulerException("This task already has a schedule. "
                                           "Create a new scheduled task.")
        self.__schedule_vsv_only_file(file_to_import, "nessus")
        self.__commit_schedule()

    def schedule_vsv_patchlink_vms(self, file_to_import):
        if self.__task_id is not None:
            raise ImpactSchedulerException("This task already has a schedule. "
                                           "Create a new scheduled task.")
        self.__schedule_vsv_only_file(file_to_import, "patchlink")
        self.__commit_schedule()

    def schedule_vsv_qualys_guard(self, file_to_import):
        if self.__task_id is not None:
            raise ImpactSchedulerException("This task already has a schedule. "
                                           "Create a new scheduled task.")
        self.__schedule_vsv_only_file(file_to_import, "qualys")
        self.__commit_schedule()

    def schedule_vsv_retina(self, file_to_import, dsn, risk_level):
        """ risk_level: string: ALL|HIGH|MEDIUM|LOW|INFORMATION """
        if self.__task_id is not None:
            raise ImpactSchedulerException("This task already has a schedule. "
                                           "Create a new scheduled task.")
        self.__schedule_vsv_only_file(file_to_import, "retina")
        self.__mod_params['DSN'] = str(dsn)
        self.__mod_params['RISKLEVEL'] = str(risk_level.uppper())
        self.__commit_schedule()

    def schedule_vsv_saint(self, file_to_import):
        if self.__task_id is not None:
            raise ImpactSchedulerException("This task already has a schedule. "
                                           "Create a new scheduled task.")
        self.__schedule_vsv_only_file(file_to_import, "saint")
        self.__commit_schedule()

    def schedule_vsv_stat_guardian(self, file_to_import):
        if self.__task_id is not None:
            raise ImpactSchedulerException("This task already has a schedule. "
                                           "Create a new scheduled task.")
        self.__schedule_vsv_only_file(file_to_import, "harris")
        self.__commit_schedule()

    def __schedule_vsv_only_file(self, file_to_import,importer):
        self.__mod_name = "Vulnerability Scanner Validator"
        self.__mod_params['IMPORTER'] = importer
        self.__mod_params['IMPORTFILENAME'] = str(file_to_import)

    def __commit_schedule(self):
        if self.__mod_name is None:
            raise ImpactSchedulerException(
                          "You haven't defined which module will be scheduled.")
        impact_com = ImpactComManager().get_impact_com()
        scheduler = impact_com.getScheduler()
        self.__task_id = scheduler.createScheduledTask(self.__name,
                                                       self.__mod_name,
                                                       self.__get_params_xml(),
                                                       self.__description)

        #create the corresponding triggers according the frequency
        # type       minute	hour day_of_month month	day_of_week	year
        # once		 x	    x	 x		      x	    x		    x
        # daily		 x	    x	 -1		      -1	-1		    x
        # workdays	 x	    x	 -1		      -1	(1-5)		x
        # mo We Fr	 x	    x	 -1		      -1	(1,3,5)		x
        # Tu Th		 x	    x	 -1		      -1	(2,4)		x
        # every week x	    x	 -1		      -1	5		    x
        # every mont x	    x	 x		      -1	-1		    x

        triggers = []
        minute, hour, day_of_month, month, day_of_week, year = (
                                                    self.__sched_time[4],
                                                    self.__sched_time[3],
                                                    self.__sched_time[2],
                                                    self.__sched_time[1],
                                                    int(time.strftime("%w",
                                                        self.__sched_time)) + 1,
                                                    self.__sched_time[0])

        if self.__frequency == "once":
            this_triger = [minute, hour, day_of_month, month, day_of_week, year]
            triggers.append(this_triger)

        elif self.__frequency == "daily":
            this_triger = [ minute, hour, -1, -1, -1, year ]
            triggers.append(this_triger)

        elif self.__frequency == "working":
            for i in range(2,7):
                this_triger = [ minute, hour, -1, -1, i, year ]
                triggers.append(this_triger)

        elif self.__frequency == "monwedfri":
            for i in range(2,7,2):
                this_triger = [ minute, hour, -1, -1, i, year ]
                triggers.append(this_triger)

        elif self.__frequency == "tuethu":
            for i in range(3,6,2):
                this_triger = [ minute, hour, -1, -1, i, year ]
                triggers.append(this_triger)

        elif self.__frequency == "weekly":
            this_triger = [ minute, hour, -1, -1, day_of_week, year ]
            triggers.append(this_triger)

        elif self.__frequency == "monthly":
            this_triger = [ minute, hour, day_of_month, -1, -1, year ]
            triggers.append(this_triger)

        for t in triggers:
            scheduler.addScheduledTaskTrigger(self.__task_id, t[0], t[1], t[2],
                                              t[3], t[4], t[5])

        #scheduler.runScheduledTask(task_id)

    def __get_params_xml(self):
        #lets use logic already implemented in impact_manager.Module
        mod = Module(None, None)
        #first add common parameters
        mod.add_parameter("EMAIL_DELIVERY","pdf","string")
        mod.add_parameter("EMAIL_REPORT","NO","string")
        mod.add_parameter("SCHEDULE_DATE",time.strftime("%Y-%m-%d",
                                                    self.__sched_time),"string")
        mod.add_parameter("SCHEDULE_FREQ",self.__frequency,"string")
        mod.add_parameter("SCHEDULE_TIME",time.strftime("%H:%M:%S",
                                                    self.__sched_time),"string")
        #add module specific parameters
        for k,v in self.__mod_params.items():
            param_type = "string"
            if type(1) == type(v):
                param_type = "int16"
            mod.add_parameter(k,str(v),param_type)
        return mod.get_params_xml(False)


class ExecutedScheduledTask:

    def __init__(self, _id, name):
        self.__id = _id
        self.__name = name

    def stop(self):
        """Attempsts to stop this task (if a schedule_x method has been called
        previously) if no schedule_x method has been called or something goes
        wrong (e.g. the task is not running) throws an ImpactModuleException."""

        if self.__id is None:
            raise ImpactModuleException("There is no scheduled task to stop")
        try:
            impact_com = ImpactComManager().get_impact_com()
            scheduler = impact_com.getScheduler()
            scheduler.stopScheduledTask(self.__id)
        except Exception, e:
            raise ImpactSchedulerException("Error stopping scheduled task %d: %s"
                                           % (self.__id, str(e)))
        except:
            raise ImpactSchedulerException(
                                   "Unexpected error stopping scheduled task %d"
                                   % self.__id)

    def get_scheduled_task_name(self):
        """Returns the name of the Scheduled task associated to this executed
        scheduled task."""
        return self.__name

    def delete(self):
        """Deletes the workspace running this scheduled task."""
        WorkspaceManager.delete_workspace_by_id(self.__id)

    def get_date_time(self):
        """Returns a struct time containing the date and time of this executed
        scheduled task"""
        name, _ = self.__get_name_status()
        s = self.__name + " %Y-%m-%d at %H:%M"
        return time.strptime(name, s)

    def open(self):
        """Returns an instance of Workspace"""
        return WorkspaceManager.open_workspace_by_id(self.__id, "")

    def get_status(self):
        """Returns 'Stopped', 'Stopping', 'Completed', or 'Running'"""
        #enum status
        #INITIALIZED = 0
        #STOPPING    = 2
        #STOPPED     = 3
        #RUNNING     = 4
        #ABORTED     = 5
        #COMPLETED   = 6
        #FINISHING   = 8
        #ABORTING    = 9
        #FINALIZED   = 10

        _, status = self.__get_name_status()
        if status is None:
            raise ImpactSchedulerException(
                         "Unable to retrieve status of executed scheduled task")
        if status in [3,5]:
            return "Stopped"
        if status in [2]:
            return "Stopping"
        if status in [6]:
            return "Completed"
        return "Running"


    def __get_name_status(self):
        db = ImpactDBManager()
        rs = db.execute("get_executed_scheduled_tasks")
        name, status = None, None
        while not rs.EOF:
            if int(self.__id) == rs.value("Workspace_Id"):
                name = rs.value("Name")
                status = rs.value("Status")
                break
            rs.next()
        return name, status