#--
#
# Author:
#   Sebastian Tello
#
# Copyright (c) 2003,2012 CORE Security Technologies, CORE SDI Inc.
# All rights reserved.
#
# This computer software is owned by Core SDI Inc. and is
# protected by U.S. copyright laws and other laws and by international
# treaties.  This computer software is furnished by CORE SDI Inc.
# pursuant to a written license agreement and may be used, copied,
# transmitted, and stored only in accordance with the terms of such
# license and with the inclusion of the above copyright notice.  This
# computer software or any other copies thereof may not be provided or
# otherwise made available to any other person.
#
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI Inc. BE LIABLE
# FOR ANY DIRECT,  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR
# CONSEQUENTIAL  DAMAGES RESULTING FROM THE USE OR MISUSE OF
# THIS SOFTWARE
#
#--

from impact_manager.impact_db_manager import ImpactDBManager
from impact_manager.entities.entity import Entity
from impact_manager.entities.agent import Agent
from impact_manager.impact_lib_exceptions import ImpactEntityException


class EntityManager:

    #ENTITY_EVENTS SUBSCRIPTIONS
    EVENT_ENTITY_CREATED = 1
    EVENT_ENTITY_UPDATED = 2
    EVENT_ENTITY_DELETED = 3

    def __init__(self, workspace_entity):
        """ creates an entity manager for a given workspace.
        The workspace parameter is a impact_manager.Workspace entity"""
        self.__workspace = workspace_entity
        self.__entity_mgr = workspace_entity.get_entity_manager_com()

    def get_entity(self, id_or_name):
        """ returns an impact_manager.Entity entity. Throws
        ImpactEntityException if something goes wrong"""
        if isinstance(id_or_name, int):
            #get by id
            if not self.entity_id_exists(id_or_name):
                raise ImpactEntityException(
                     "Entity with id %d does not exists in workspace %s(%d)"
                     % (id_or_name, self.__workspace.name, self.__workspace.id))
            return Entity(self.__workspace, id_or_name)
        else:
            id_or_name = str(id_or_name)
            if not self.entity_exists(id_or_name):
                raise ImpactEntityException(
                     "Entity '%s' does not exists in workspace %s(%d)"
                     % (id_or_name, self.__workspace.name, self.__workspace.id))

            try:
                #TODO: USE getEntityByName(name) to get the real entity name by
                #a display name
                ent_id = self.get_entity_id_by_name(id_or_name)
            except Exception:
                raise ImpactEntityException(
                     "Entity '%s' does not exists in workspace %s(%d)"
                     % (id_or_name, self.__workspace.name, self.__workspace.id))
            return Entity(self.__workspace, ent_id)

    def get_entity_named_like(self, likename):
        """ Similar to get_enity. But only accepts an string, and the entity
        returned is the first entity found whose name matches likename.
        Throws an ImpactEntityException if something goes wrong"""
        entity_list = self.list_entities()
        for en in entity_list:
            if likename in en:
                return self.get_entity(en)
        raise ImpactEntityException(
                       "Entity named like '%s' wasn't found in workspace %s(%d)"
                       % (likename, self.__workspace.name, self.__workspace.id))

    def get_agent(self, name):
        """ returns an impact_manager.entities.agent.Agent entity representing
        an agent that matches with name. If no agent matching name is found an
        ImpactEntityException is raised """
        found = False
        ag_name = ""
        for ag in self.list_agents():
            if name in ag:
                ag_name = ag
                found = True
                break
        if not found:
                raise ImpactEntityException(
                           "Agent '%s' does not exists in workspace %s(%d)"
                           % (name, self.__workspace.name, self.__workspace.id))
        ag_id = self.get_entity_id_by_name(ag_name)
        return Agent(self.__workspace, ag_id)

    def get_agent_named_like(self, likename):
        """ Similar to get_agent. But only accepts an string, and the agent
        returned is the first agent found whose name matches likename.
        If no agent matching name is found an ImpactEntityException is raised"""

        #get agent's complete name
        agent_list = self.list_agents()
        for ag in agent_list:
            if likename in ag:
                return self.get_agent(ag)
        raise ImpactEntityException(
                        "Agent named like '%s' wasn't found in workspace %s(%d)"
                       % (likename, self.__workspace.name, self.__workspace.id))


    def get_entity_id_by_name(self, name):
        """ Returns an integer with the id of the entity whit the given name.
        Raises an ImpactEntityException if the entity does not exist in the
        current workspace"""
        if not self.entity_exists(name):
            raise ImpactEntityException(
                           "Entity '%s' does not exists in workspace %s(%d)"
                           % (name, self.__workspace.name, self.__workspace.id))
        db = ImpactDBManager()
        query="SELECT Id FROM TEntity WHERE Name = '%s' and Workspace_Id = %d;"\
              % (db.addslashes(name), self.__workspace.id)
        rs = db.execute(query)
        return int(rs.value(0))

    def get_webapps_site_entity_name(self, scenario_display_name,
                                     site_display_Name):
        """ This method transform a display_name of a site into a hash, but also
        you need to pass the scenario name. Receives the scenario_display_name,
        and the site_display_name and returns the hash name or the site.
        This method is usefull when you are working with unique scenarios name
        in the same workspace and you have the same page in the different
        scenarios.
        """
        scenarioURLs =self.list_webapps_entities_by_owner(scenario_display_name,
                                                          returnHash=True)
        targetUrlHash = ""
        for URL in scenarioURLs:
            if ( site_display_Name == self.get_webapps_display_name(URL)):
                targetUrlHash = URL
        return targetUrlHash

    def get_webapps_agent_entity_name(self, scenario_display_name,
                                                             site_display_Name):
        """This method gets all the webapps agents that are in a specified page.
        You need to provide a display_name of a site, and scenario name.
        Receives the scenario_display_name, and the site_display_name and
        returns a list of tuples with (hash names, display name) or all the
        webapps agent in the site. This method is usefull when you are working
        with unique scenarios name in the same workspace and you have the same
        page in the different scenarios.
        """
        toReturn = []
        scenarioURLs =self.list_webapps_entities_by_owner(scenario_display_name,
                                                          returnHash=True)
        targetUrlHash = ""
        for URL in scenarioURLs:
            if ( site_display_Name == self.get_webapps_display_name(URL)):
                targetUrlHash = URL

        hashAgentsList = self.list_webapps_entities_by_hash_owner(targetUrlHash,
                                                                returnHash=True)
        for i in hashAgentsList:
            toReturn.append((i,self.get_webapps_display_name(i)))
        return toReturn

    def get_webapps_entity_name(self, display_name):
        """DEPRECATED"""
        import warnings
        from exceptions import DeprecationWarning
        warnings.warn("get_webapps_entity_name is DEPRECATED you should use " \
                      "get_entity_name(display_name) instead",
                      DeprecationWarning)

        return self.get_entity_name(display_name)

    def get_webapps_display_name(self, entity_name):
        """DEPRECATED"""
        import warnings
        from exceptions import DeprecationWarning
        warnings.warn("get_webapps_display_name is DEPRECATED you should use "\
                      "get_entity_display_name(entity_name) instead",
                      DeprecationWarning)

        return self.get_entity_display_name(entity_name)

    def get_entity_name(self, display_name, entity_class=None):
        """ Get an entity name by it's display name.
        entity_class might be:
            None: don't filter by entity class
            string: filter by the given entity class
            list os strings: filter by the given entity classes
        Warning: This method is reliable only works when the display name is
        unique in the Workspace.
        """
        #Even though impact shows a pretty name when working with entities
        #(e.g Hosts, Emails, XSS Agents, Scenarios, Web Pages, Browsers, etc)
        #Behind the curtains what really matters is the entity_name that can be
        #anything.

        #TODO: Probably using getEntityByName(name) to get the real entity name
        #by a display name
        db = ImpactDBManager()
        class_filter = ""
        if entity_class:
            if not isinstance(entity_class, list):
                entity_class = [str(entity_class)]

            class_filter = "AND E.Class IN (%s) "  % ','.join(
                "'%s'" % (db.addslashes(c)) for c in entity_class)

        query = "SELECT E.Name " \
                "FROM TEntity E " \
                "JOIN TEntityProperty EP " \
                "ON E.Id = EP.Entity_Id " \
                "AND E.Workspace_Id = EP.Workspace_Id " \
                "WHERE E.Workspace_Id = %d " \
                "%s" \
                "AND EP.Name = 'display_name' " \
                "AND EP.Value_Tx = '%s';" % (self.__workspace.id,
                                             class_filter,
                                             db.addslashes(display_name))
        rs = db.execute(query)
        return rs.value(0)

    def get_entity_display_name(self, entity_name):
        """ Even though impact shows a pretty name when working with entities
        (e.g Hosts, Emails, XSS Agents, Scenarios, Web Pages, Browser info, etc)
        Behind the curtains what really matters is the entity_name that could be
        anything.
        This method return the display name when you have the entity_name
        This can be used for verification pourposes
        """
        db = ImpactDBManager()
        query = "SELECT  EP.Value_Tx " \
                "FROM TEntity E " \
                "JOIN TEntityProperty EP " \
                "ON E.Id = EP.Entity_Id " \
                "AND E.Workspace_Id = EP.Workspace_Id " \
                "WHERE E.Workspace_Id = %d " \
                "AND EP.Name = 'display_name' " \
                "AND E.Name = '%s';" % (self.__workspace.id, entity_name)
        rs = db.execute(query)
        return rs.value(0)

    def delete_entity_by_id(self, id):
        """ Deletes an entity from the workspace by its id
        Returns True if the entity was deleted succesfully. False otherwise. """
        try:
            self.__entity_mgr.eraseEntity(id)
        except:
            return False
        return True

    def delete_entity_by_name(self, name):
        """ Deletes an entity from the workspace by its name
        Returns True if the entity was deleted succesfully. False otherwise. """
        try:
            self.__entity_mgr.eraseEntityByName(name)
        except:
            return False
        return True

    def delete_entities_named_like(self, likename):
        """Delete all entities whose name match likename"""
        while True:
            try:
                entity = self.get_entity_named_like(likename)
                self.delete_entity_by_id(entity.get_id())
            except:
                break

    def entity_id_exists(self, id):
        """ Returns True if an entity with this id exists in the workspace """
        try:
            self.__entity_mgr.getEntity(id)
        except:
            return False
        return True

    def entity_exists(self, name):
        """ Returns True if an entity with this name exists in the workspace """
        try:
            self.__entity_mgr.getEntityByName(name)
        except:
            return False
        return True

    def list_hosts(self):
        """ Returns a list of host names in the current workspace """
        return self.list_entities_by_class("host")

    def list_access_points(self):
        """ Returns a list of Access Points names in the current workspace """
        return self.list_entities_by_class("AccessPoint")

    def list_agents(self):
        """ Returns a list of agent names in the current workspace """
        return self.list_entities_by_class("agent")

    def list_entities(self):
        """ Returns a list of all the entity names in the current workspace """
        db = ImpactDBManager()
        query = "SELECT Name FROM TEntity WHERE Workspace_Id = %d;" % (
                                                            self.__workspace.id)
        rs = db.execute(query)
        ret = []
        while not rs.EOF:
            ret.append(rs.value(0))
            rs.next()
        return ret

    def list_entities_by_class(self, class_name):
        """ Returns a list of entity names that belong to the given class name
        for the current workspace"""
        db = ImpactDBManager()
        class_name = db.addslashes(str(class_name))
        query = "SELECT Name FROM TEntity WHERE Class = '%s' and " \
                "Workspace_Id = %d;" % (class_name, self.__workspace.id)
        rs = db.execute(query)
        ret = []
        while not rs.EOF:
            ret.append(rs.value(0))
            rs.next()
        return ret

    def list_webapps_entities_by_owner(self, ownerDisplayName,returnHash=False):
        """ Returns a list of entity names that have owner the given entity name
        for the current workspace"""
        db = ImpactDBManager()
        query = "SELECT E.Name " \
                "FROM TEntity E " \
                "JOIN TEntityProperty EP " \
                "ON E.Id = EP.Entity_Id " \
                "AND E.Workspace_Id = EP.Workspace_Id " \
                "WHERE E.Workspace_Id = %d " \
                "AND EP.Name = 'owner' " \
                "AND EP.Value_Tx = '%s';" % (self.__workspace.id,
                                 self.get_webapps_entity_name(ownerDisplayName))
        rs = db.execute(query)
        ret = []
        while not rs.EOF:
            if (returnHash):
                ret.append(rs.value(0))
            else:
                ret.append(self.get_webapps_display_name(rs.value(0)))
            rs.next()
        return ret

    def list_webapps_entities_by_hash_owner(self, ownerHashName,
                                            returnHash=False):
        """ Returns a list of entity names that have owner the given entity name
        for the current workspace"""
        if not ownerHashName or ownerHashName == "":
            return []
        db = ImpactDBManager()
        query = "SELECT E.Name " \
                "FROM TEntity E " \
                "JOIN TEntityProperty EP " \
                "ON E.Id = EP.Entity_Id " \
                "AND E.Workspace_Id = EP.Workspace_Id " \
                "WHERE E.Workspace_Id = %d " \
                "AND EP.Name = 'owner' " \
                "AND EP.Value_Tx = '%s';" % (self.__workspace.id, ownerHashName)
        rs = db.execute(query)
        ret = []
        while not rs.EOF:
            if (returnHash):
                ret.append(rs.value(0))
            else:
                ret.append(self.get_webapps_display_name(rs.value(0)))
            rs.next()
        return ret

    def subscribe_to_entity_events(self, event_types, function,
                                                        *additional_parameters):
        """Subscribe a function to events related to entities in this workspace.
        event_types: is a list of event types you want to be notified about
            e.g. [EntityManager.EVENT_ENTITY_UPDATED,
                  EntityManager.EVENT_ENTITY_DELETED]
        function: the function that will be called when the event occurs
        additional_parameters: When the callback function is called, it will
            provide a dictionary parameter containing information about the
            event, you can give any other parameters you want to be passed
            besides the data dictionary

        The data dictionary passed back to the subscribed functions contains:
            'EVENT_TIME': float: number of seconds since the epoc
            'EVENT_TYPE': int: either EntityManager.EVENT_ENTITY_CREATED
                                      EntityManager.EVENT_ENTITY_UPDATED,
                                      or EntityManager.EVENT_ENTITY_DELETED
            'WORKSPACE_ID': int: workspace_id,
            'TASK_ID': int: task_id,
            'ENTITY_ID': int: entity_id
            'ENTITY_NAME': string: name of the entity
            'ENTITY_CLASS': string: class of the entity

        Note: Keep in mind that you won't get event notifications related to
        entities in other workspaces.
        Example of usage:
        def myfunc(data, greeting):
            print greeting
            print "your event data is:", data
        emgr = EntityManager(wsp)
        which_events = [EntityManager.EVENT_ENTITY_UPDATED,
                       EntityManager.EVENT_ENTITY_CREATED]
        emgr.subscribe_to_entity_events(which_events, myfunc, "Hi There!")
        """
        entity_mgr_event_handler=self.__workspace.get_entity_mgr_event_handler()
        entity_mgr_event_handler.subscribe_by_workspace(event_types, function,
                                                          additional_parameters)
        
    def __repr__(self):
        return "<EntityManager on Workspace:%d>" % self.__workspace.id
