#--
#
# Author:
#   Sebastian Tello
#
# Copyright (c) 2003,2012 CORE Security Technologies, CORE SDI Inc.
# All rights reserved.
#
# This computer software is owned by Core SDI Inc. and is
# protected by U.S. copyright laws and other laws and by international
# treaties.  This computer software is furnished by CORE SDI Inc.
# pursuant to a written license agreement and may be used, copied,
# transmitted, and stored only in accordance with the terms of such
# license and with the inclusion of the above copyright notice.  This
# computer software or any other copies thereof may not be provided or
# otherwise made available to any other person.
#
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI Inc. BE LIABLE
# FOR ANY DIRECT,  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR
# CONSEQUENTIAL  DAMAGES RESULTING FROM THE USE OR MISUSE OF
# THIS SOFTWARE
#
#--

import comtypes.client

from impact_manager.resources.singleton import Singleton
from impact_manager.resources.wservice import WService
from impact_manager.resources.entity_mgr_event_handler \
                                                    import EntityMgrEventHandler
from impact_manager.resources.general_module_execution_event_handler \
                                       import GeneralModuleExecutionEventHandler
from impact_db_manager import ImpactDBManager

class ImpactComManager:
    __metaclass__ = Singleton

    def __init__(self):
        """This is a singleton object. Only the first call to the constructor
        will instantiate the object, following calls will just return the same
        instance."""
        self.__service_name = "CORE IMPACT Service"
        self.__log = []
        self.start()
        self.get_impact_com()

    def __del__(self):
        """Close the com objects before leaving"""
        try:
            del self.__workspace_manager
            del self.__module_manager
            del self.__license_manager
            del self.__update_manager
            self.shutdown()
            del self.__impact
        except:
            pass


    def get_impact_com(self):
        """Returns the main impact COM object. First checks if it is up
        if not tries to start it"""
        if not self.is_up():
            dbmgr = ImpactDBManager()
            dbmgr.check_connection_is_opened()
            self.stop_service()
            self.start()
        return self.__impact

    def get_workspace_mgr_com(self):
        """Returns the impact's WorkspaceMgr COM object"""
        #just to make sure the impact com is up
        #this call will test it up, and regenerates the com objects if they were
        #killed
        self.get_impact_com()
        return self.__workspace_manager

    def get_module_mgr_com(self):
        """Returns the impact's ModuleMgr COM object"""
        #just to make sure the impact com is up
        #this call will test it up, and regenerates the com objects if they were
        #killed
        self.get_impact_com()
        return self.__module_manager

    def get_license_mgr_com(self):
        """Returns the impact's LicenseMgr COM object"""
        #just to make sure the impact com is up
        #this call will test it up, and regenerates the com objects if they were
        #killed
        self.get_impact_com()
        return self.__license_manager

    def get_update_mgr_com(self):
        """Returns the impact's UpdateMgr COM object"""
        self.get_impact_com()
        return self.__update_manager

    @staticmethod
    def get_entity_mgr_event_handler():
        """Instances and EntityMgrEventHanlder COM interface"""
        #Returns an instance of an entity manager event handler
        return EntityMgrEventHandler()

    @staticmethod
    def get_general_module_execution_event_handler():
        """Instances and ModuleExecutionEventHanlder COM interface"""
        return GeneralModuleExecutionEventHandler()

    def is_up(self):
        """Runs a few checks in the impact com, returns True if it is up,
        False otherwise"""
        try:
            #we will run actions that will throw exeptions if the services are
            #down
            log = self.__impact.getSystemLog()
            self.__log = log.getLines()
            wspmgr = self.__impact.getWorkspaceMgr()
            wspmgr.getWorkspaceId("zaraza")
            return True
        except:
            return False

    def start(self):
        """Starts the Impact's COM service"""
        self.start_service()

        #start the main impact object
        self.__impact = comtypes.client.CreateObject("impact_core_com.Impact")
        self.__impact.initialize()

        #get the workspace manager and the module managers
        self.__workspace_manager = self.__impact.getWorkspaceMgr()
        self.__module_manager = self.__impact.getModuleMgr()
        self.__license_manager = self.__impact.getLicenseMgr()
        self.__update_manager = self.__impact.getUpdateMgr()

    def shutdown(self):
        """If up, shuts down the instance of the impact COM object"""
        if self.is_up():
            #commented out. See sniper/ticket/19863
            #self.__impact.shutdown()
            pass

    def last_known_log(self):
        """Returns the last known Impact System Log"""
        self.is_up()
        return self.__log

    def start_service(self):
        """ Attemps to start the Core Impact service. Returns True if it is
        running (whether it was started or it was already running). Otherwise
        returns False """
        try:
            impact = WService(self.__service_name)
            if impact.status() is not "RUNNING":
                impact.start()
                #wait up to 20 seconds for the service to start
                result = impact.fetchstatus("RUNNING", 60)
                if result is "TIMEDOUT":
                    return False
        except:
            return False
        return True

    def stop_service(self):
        """ Attemps to stop the Impact service. Returns True if it is not
        running (whether it was stopped or it was already shutted down).
        Otherwise returns False """
        try:
            impact = WService(self.__service_name)
            if impact.status() is not "STOPPED":
                impact.stop()
                #wait up to 60 seconds for the service to stopped
                result = impact.fetchstatus("STOPPED", 60)
                if result is "TIMEDOUT":
                    return False
        except Exception, e:
            print e
            return False
        return True

    def restart_service(self):
        """ Calls stop service, and then start service. Returns the status of
        start service"""
        self.stop_service()
        return self.start_service()


    @staticmethod
    def get_guid():
        """Returns Impact's COM guid"""
        return "{ED4FFED7-CB5F-429C-8ED3-A3BD8EC84249}"
