#--
#
# Author:
#   Sebastian Tello
#
# Copyright (c) 2003,2012 CORE Security Technologies, CORE SDI Inc.
# All rights reserved.
#
# This computer software is owned by Core SDI Inc. and is
# protected by U.S. copyright laws and other laws and by international
# treaties.  This computer software is furnished by CORE SDI Inc.
# pursuant to a written license agreement and may be used, copied,
# transmitted, and stored only in accordance with the terms of such
# license and with the inclusion of the above copyright notice.  This
# computer software or any other copies thereof may not be provided or
# otherwise made available to any other person.
#
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI Inc. BE LIABLE
# FOR ANY DIRECT,  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR
# CONSEQUENTIAL  DAMAGES RESULTING FROM THE USE OR MISUSE OF
# THIS SOFTWARE
#
#--

import sys
import os
import shutil
from xml.dom import minidom

from impact_manager.resources.singleton import Singleton
from impact_manager.resources import shutil_patched
from impact_manager.impact_com_manager import ImpactComManager
from impact_manager.impact_configuration_manager \
                                               import ImpactConfigurationManager
from impact_manager.impact_db_manager import ImpactDBManager
from impact_manager.impact_lib_exceptions import ImpactModuleException

class ModuleManager:
    __metaclass__ = Singleton

    def __init__(self):
        """ Handles general modules operation in Impact.
        All the methods in this class are static.
        You dont need an instance."""
        pass

    @staticmethod
    def get_module_list_by_category(category_like):
        """ Returns a list of modules whose categories contains the string
        category_like. The search is case sensitive so 'Remote' is not the same
        than 'remote' """

        db = ImpactDBManager()
        category_like = db.addslashes(category_like)
        category_like = category_like.replace('%', '\\%')

        query = "SELECT ModuleName FROM TModule WHERE Category like '%%%s%%';" \
                                                               % (category_like)
        rs = db.execute(query)
        ret = []
        while not rs.EOF:
            ret.append(rs.value('ModuleName'))
            rs.next()
        return ret

    @staticmethod
    def get_distinct_module_paths():
        """ Returns the list of files containing modules imported by Impact """
        db = ImpactDBManager()

        query = "SELECT DISTINCT(M.Path) FROM TModuleFile as M"
        rs = db.execute(query)
        ret = []

        while not rs.EOF:
            ret.append(rs.value('Path'))
            rs.next()
        return ret

    @staticmethod
    def install_module(file_dir, override=True, recurse_flag=None):
        """ Installs the given file or directory. In impact's User Modules
        directory. If the files already existed and override is False, the 
        original files are not modified.
        If recurse_flag is None (default), the recurse registry setting is not
        checked. If True or False, the method checks if the flag is ON or OFF
        respectively, if it's different it's changed and the COM service is
        restarted.
        """
        try:
            file_dir = os.path.normpath(file_dir)
            user_mods= ImpactConfigurationManager.get_registry_value("Paths", 
                                                            "PythonUserModules")

            dst = os.path.join(user_mods, os.path.basename(file_dir))
            src = file_dir

            if isinstance(recurse_flag, bool):
                ModuleManager.__set_user_modules_recurse(recurse_flag)

            #If the destination path or file already exists
            #and we don't want to overwrite
            if os.path.exists(dst) and not override:
                return True

            if os.path.isfile(src):
                directory = os.path.dirname(file_dir)
                if directory:
                    if os.path.exists(dst):
                        os.remove(dst)
                    if not os.path.exists(os.path.dirname(dst)):
                        os.makedirs(os.path.dirname(dst), mode=0777)
                shutil.copy(src, dst)
                ModuleManager.reload_modules()
                return True
            
            elif os.path.isdir(src):
                if os.path.exists(dst):
                    shutil.rmtree(dst)
                if sys.version_info < (2, 6):
                    shutil_patched.copytree(src,dst,
                        ignore=shutil_patched.ignore_patterns('*.svn', '*.pyc'))
                else:
                    shutil.copytree(src,dst,
                        ignore=shutil_patched.ignore_patterns('*.svn', '*.pyc'))
                ModuleManager.reload_modules()
                return True
            return False
        except Exception, e:
            raise ImpactModuleException(
                                    "Error when installing module: %s" % str(e))

    @staticmethod
    def stop_running_modules(workspace):
        """Attemps to stop all the running modules from the given workspace.
        workspace is an impact_manager.Workspace object.
        Throws an ImpactModuleException if something goes wrong """
        try:
            me = workspace.get_module_execution_com()
            me.stopAll()
        except Exception, e:
            raise ImpactModuleException(
                               "Error when trying to stop modules: %s" % str(e))

    @staticmethod
    def reload_modules():
        """ Forces the impact COM service to reload modules. Same effect than
        executing Modules->Reload from the GUI.
        Throws an ImpactModuleException if something goes wrong """
        try:
            icm = ImpactComManager().get_impact_com()
            mod_mgr = icm.getModuleMgr()
            mod_mgr.reloadModules()
        except Exception, e:
            raise ImpactModuleException(
                                    "Error when reloading modules: %s" % str(e))

    @staticmethod
    def get_module_property(module_name, prop_name):
        """ Gets a property from the Module's XML. If the property is composed
        of more properties, returns a dictionary with the keys and values of
        the nested properties """
        db = ImpactDBManager()
        module_name = db.addslashes(module_name)
        query = "SELECT XML FROM TModule WHERE ModuleName = '%s'" %(module_name)
        rs = db.execute(query)

        xml = rs.value('XML')

        try:
            doc = minidom.parseString(xml)
        except:
            doc = minidom.parseString(xml.encode("UTF-8"))
        root = doc.getElementsByTagName("entity")[0]
        for prop in root.getElementsByTagName("property"):
            if prop.hasAttribute("key") and prop.getAttribute("key")==prop_name:
                return ModuleManager.__get_data_from_property(prop)
        return None


    @staticmethod
    def __set_user_modules_recurse(recurse):
        """Checks/sets the user modules recurse flag, if changed restarts the
        COM service"""
        
        reg_path = r"\Paths\PythonModules\User"
        reg_recurse = ImpactConfigurationManager.get_registry_value(reg_path, 
                                                                    "Recurse")
        
        #if the flag is the one we want, do nothing
        if not (recurse or reg_recurse) or (recurse and reg_recurse):
            return
        
        new_val = 0
        if recurse:
            new_val = 1
            
        ImpactConfigurationManager._set_registry_value(reg_path, "Recurse",
                                                       "REG_DWORD", new_val)
        #Restart impact_core_com
        icm = ImpactComManager()
        icm.shutdown()
        icm.restart_service()
        icm.start()    


    @staticmethod
    def __get_data_from_property(parent):
        if (len(parent.childNodes) == 1
                      and parent.firstChild.nodeType == minidom.Node.TEXT_NODE):
            return parent.firstChild.nodeValue
        elif len(parent.childNodes) == 0:
            return ""
        else:
            ret = {}
            for child in parent.childNodes:
                if child.nodeName == "property" and child.hasAttribute("key"):
                    ret[child.getAttribute("key")
                        ] = ModuleManager.__get_data_from_property(child)
            return ret