#--
#
# Author:
#   Sebastian Tello
#
# Copyright (c) 2003,2012 CORE Security Technologies, CORE SDI Inc.
# All rights reserved.
#
# This computer software is owned by Core SDI Inc. and is
# protected by U.S. copyright laws and other laws and by international
# treaties.  This computer software is furnished by CORE SDI Inc.
# pursuant to a written license agreement and may be used, copied,
# transmitted, and stored only in accordance with the terms of such
# license and with the inclusion of the above copyright notice.  This
# computer software or any other copies thereof may not be provided or
# otherwise made available to any other person.
#
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI Inc. BE LIABLE
# FOR ANY DIRECT,  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR
# CONSEQUENTIAL  DAMAGES RESULTING FROM THE USE OR MISUSE OF
# THIS SOFTWARE
#
#--

import time
import comtypes

from comtypes.gen import impact_core_comLib as impact_core_com
from impact_manager.impact_db_manager import ImpactDBManager

class EntityMgrEventHandler(comtypes.COMObject):
    _com_interfaces_ = [impact_core_com.IEntityMgrEventHandler]

    #EVENT TYPES
    ENTITY_CREATED = 1
    ENTITY_UPDATED = 2
    ENTITY_DELETED = 3

    def __init__(self):
        self.__listeners_by_task = {EntityMgrEventHandler.ENTITY_CREATED:[],
                                    EntityMgrEventHandler.ENTITY_DELETED:[],
                                    EntityMgrEventHandler.ENTITY_UPDATED:[]}
        #id(task), callback_fc, *callback_params

        self.__listeners_by_task_tree ={EntityMgrEventHandler.ENTITY_CREATED:[],
                                        EntityMgrEventHandler.ENTITY_DELETED:[],
                                        EntityMgrEventHandler.ENTITY_UPDATED:[]}
        #id(main_task), callback_fc, *callback_params

        self.__listeners_by_entity = {EntityMgrEventHandler.ENTITY_CREATED:[],
                                      EntityMgrEventHandler.ENTITY_DELETED:[],
                                      EntityMgrEventHandler.ENTITY_UPDATED:[]}
        #id (entity), callback_fc, *callback_params

        self.__listeners_by_workspace ={EntityMgrEventHandler.ENTITY_CREATED:[],
                                        EntityMgrEventHandler.ENTITY_DELETED:[],
                                        EntityMgrEventHandler.ENTITY_UPDATED:[]}
        #id (-1), callback_fc, *callback_params

    def wrap_me(self):
        """"Returns a wrapper of this interface to be passed as argument to the
        COM function"""
        return self.QueryInterface(impact_core_com.IEntityMgrEventHandler)

    def subscribe_by_entity(self, entity_id, event_types, callback_fc,
                            *callback_additional_params):
        """Subscribes a function to the callback list for the required
        event_types on the entity with id entity_id"""
        listener = {}
        listener['id'] = entity_id
        listener['callback_fc'] = callback_fc
        listener['callback_params'] = callback_additional_params

        for event_type in event_types:
            if self.__listeners_by_entity.has_key(event_type):
                self.__listeners_by_entity[event_type].append(listener)

    def subscribe_by_task(self, task_id, event_types, callback_fc,
                          *callback_additional_params):
        """Subscribes a function to the callback list for the required entity
        event_types generated by the task task_id"""
        listener = {}
        listener['id'] = task_id
        listener['callback_fc'] = callback_fc
        listener['callback_params'] = callback_additional_params

        for event_type in event_types:
            if self.__listeners_by_task.has_key(event_type):
                self.__listeners_by_task[event_type].append(listener)

    def subscribe_by_task_or_descendants(self, task_id, event_types,
                                         callback_fc,
                                         *callback_additional_params):
        """Subscribes a function to the callback list for the required entity
        event_types generated by the task task_id or any of its descendant
        tasks"""
        listener = {}
        listener['id'] = task_id
        listener['callback_fc'] = callback_fc
        listener['callback_params'] = callback_additional_params

        for event_type in event_types:
            if self.__listeners_by_task_tree.has_key(event_type):
                self.__listeners_by_task_tree[event_type].append(listener)


    def subscribe_by_workspace(self, event_types, callback_fc,
                               *callback_additional_params):
        """Subscribes a function to the callback list for the required entity
        event_types on the current workspace"""
        listener = {}
        #workspace_id not necessary as this handler is instanced for a workspace
        listener['id'] = -1
        listener['callback_fc'] = callback_fc
        listener['callback_params'] = callback_additional_params

        for event_type in event_types:
            if self.__listeners_by_workspace.has_key(event_type):
                self.__listeners_by_workspace[event_type].append(listener)


    def IEntityMgrEventHandler_onEvent(self, event_type, workspace_id, task_id,
                                       entity_id, more):
        data_dict = {'EVENT_TIME': time.time(),
                     'EVENT_TYPE':event_type,
                     'WORKSPACE_ID':workspace_id,
                     'TASK_ID':task_id,
                     'ENTITY_ID':entity_id
                    }
        for item in more:
            for subitem in item:
                data_dict[str(subitem.Key)] = str(subitem.Value)

        listeners = []
        #add listeners subscribed to this event and task_id
        for listener in self.__listeners_by_task[event_type]:
            if listener['id'] == task_id:
                listeners.append(listener)

        #add listeners subscribed to this event and task_id ir ancestor
        try:
            tree = self.__get_task_tree(workspace_id)
            for listener in self.__listeners_by_task_tree[event_type]:
                if self.__is_descendant_task(task_id, listener['id'], tree):
                    listeners.append(listener)
        except Exception, e:
            print "[EntityMgrEventHandler]Error determining Task Tree:", e

        #add listeners subscribed to this event and entity_id
        for listener in self.__listeners_by_entity[event_type]:
            if listener['id'] == entity_id:
                listeners.append(listener)

        #add listeners subscribed to any event of this type in the workspace
        for listener in self.__listeners_by_workspace[event_type]:
            #no check for workspace id is required, as this handler is instanced
            #for a workspace
            listeners.append(listener)

        for l in listeners:
            try:
                callf = l['callback_fc']
                callparams = l['callback_params'][0].__add__((data_dict,))
                callf(*callparams)
            except Exception, e:
                #TODO: what should we do if the function call, or the function
                #itself fails? Is not our fault, but the listener's. Anyway
                #we should let him know somehow
                print "[EntityMgrEventHandler]Error in callback subscriber:", e


    def __get_task_tree(self, workspace_id):
        db = ImpactDBManager()
        q = "SELECT Task_Id, ParentTask_Id " \
            "FROM TTaskRelation " \
            "WHERE ParentTask_Id != -1 " \
            "AND Workspace_Id = %d" % (workspace_id)
        rs = db.execute(q)
        tree = {}
        while not rs.EOF:
            tree[rs.value('Task_Id')] = rs.value('ParentTask_Id')
            rs.next()
        return tree

    def __is_descendant_task(self, event_task, task_id, tree):
        if event_task == task_id: return True
        cur_task = event_task
        while cur_task is not None:
            cur_task = tree.get(cur_task)
            if cur_task == task_id:
                return True

        return False