#--
#
# Author:
#   Sebastian Tello
#
# Copyright (c) 2003,2012 CORE Security Technologies, CORE SDI Inc.
# All rights reserved.
#
# This computer software is owned by Core SDI Inc. and is
# protected by U.S. copyright laws and other laws and by international
# treaties.  This computer software is furnished by CORE SDI Inc.
# pursuant to a written license agreement and may be used, copied,
# transmitted, and stored only in accordance with the terms of such
# license and with the inclusion of the above copyright notice.  This
# computer software or any other copies thereof may not be provided or
# otherwise made available to any other person.
#
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI Inc. BE LIABLE
# FOR ANY DIRECT,  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR
# CONSEQUENTIAL  DAMAGES RESULTING FROM THE USE OR MISUSE OF
# THIS SOFTWARE
#
#--

import os
import time
import subprocess

class GenLic:

    def __init__(self, genlic_path, start_date=None, expiration_date=None):
        """Wrapps the impact license generator tool. Requires the path to the
        .exe file
        if start_date is not given, today's date is used.
        if expiration_date is not given, a year ahead from start_date is used.
        dates format are: YYYY/MM/DD"""
        #Will throw an exception if the file does not exists
        os.stat(genlic_path)

        date_format = "%Y/%m/%d"
        if not start_date:
            start_date = time.strftime(date_format)
        #Next expression will throw an execption if date format is wrong
        time.strptime(start_date, date_format)

        if not expiration_date:
            now = time.strptime(start_date, date_format)
            next_year = list(now)
            next_year[0] += 1
            expiration_date = time.strftime(date_format, tuple(next_year))
        #Next expression will throw an execption if date format is wrong
        time.strptime(expiration_date, date_format)

        self._exe = genlic_path
        self._start_date = start_date
        self._expiration_date = expiration_date
        #additional options
        self._workspace = None
        self._max_host_count = None
        self._module_name = None
        self._ip_ranges = []
        self._ip_count = None
        self._warning1 = None
        self._warning2 = None
        self._max_changes = None
        self._seed = None

    def generate(self):
        """Generates the license file with the given options, returns the full
        path to the just created .lic file"""
        args = [self._exe, self._start_date, self._expiration_date]
        if self._workspace:
            args.extend(["-W", self._workspace])
        if self._max_host_count:
            args.extend(["-H", self._max_host_count])
        if self._module_name:
            args.extend(["-M", self._module_name])
        if len(self._ip_ranges) > 0:
            args.extend(["-R", ";".join(self._ip_ranges)])
        if self._ip_count:
            args.extend(["-I", self._ip_count])
        if self._warning1:
            args.extend(["-WL1", self._warning1])
        if self._warning2:
            args.extend(["-WL2", self._warning2])
        if self._max_changes:
            args.extend(["-MR", self._max_changes])
        if self._seed:
            args.extend(["-S", self._seed])

        proc = subprocess.Popen(args=args, stdout=subprocess.PIPE)
        out = proc.stdout.read()
        if "Save file in" in out:
            return os.path.abspath(" ".join(out.split()[3:]))
        else:
            return None

    def set_options(self, W=None,
                          H=None,
                          M=None,
                          R=None,
                          I=None,
                          WL1=None,
                          WL2=None,
                          MR=None,
                          S=None):
        """Sets all options (Default None, unset the option"""
        self.set_workspace_name(W)
        self.set_max_host_count(H)
        self.set_module_name(M)
        self.add_ip_range(R)
        self.set_ip_count(I)
        self.set_warning_1_trigger(WL1)
        self.set_warning_2_trigger(WL2)
        self.set_max_changes(MR)
        self.set_seed(S)

    def set_workspace_name(self, workspace_name):
        """To generate a workspace license type (genlic option -W)
        to unset the workspace license type pass None as workspace_name"""
        self._workspace = workspace_name

    def set_max_host_count(self, max_host_count):
        """Specified number of hosts permitted in this license
        (genlic option -H) to unset this option set max_host_count to None"""
        if not max_host_count:
            self._max_host_count = None
        else:
            self._max_host_count = str(int(max_host_count))

    def set_module_name(self, module_name):
        """Specified module name permitted in this license (genlic option -M)
        to unset this option set module_name to None"""
        self._module_name = module_name

    def add_ip_range(self, ip_range):
        """Add a ip range to list of ip permitted by the license
        (genlic option -R). ip_range can be:
        list: list of single ranges string
        string: a range of the type '192.168.0.0/16'
        None: Unsets the ip range license type."""
        if not ip_range:
            self._ip_ranges = []
        elif isinstance(ip_range, list):
            self._ip_ranges.extend(ip_range)
        else:
            self._ip_ranges.append(ip_range)

    def set_ip_count(self, ip_count):
        """Specified number of IPs permitted in this IPCount license
        (genlic option -I) to unset this option set ip_count to None"""
        if not ip_count:
            self._ip_count = None
        else:
            self._ip_count = str(int(ip_count))

    def set_warning_1_trigger(self, ip_count):
        """Specify the number of IPs added in IPCount to trigger warning 1
        (genlic option -WL1) to unset this option set ip_count to None"""
        if not ip_count:
            self._warning1 = None
        else:
            self._warning1 = str(int(ip_count))

    def set_warning_2_trigger(self, ip_count):
        """Specify the number of IPs added in IPCount to trigger warning 2
        (genlic option -WL2) to unset this option set ip_count to None"""
        if not ip_count:
            self._warning2 = None
        else:
            self._warning2 = str(int(ip_count))

    def set_max_changes(self, max_changes):
        """Specify the max number of changes allowed in IPCount
        (genlic option -MR) to unset this option set max_changes to None"""
        if not max_changes:
            self._max_changes = None
        else:
            self._max_changes = str(int(max_changes))

    def set_seed(self, seed):
        """Specify the license seed (genlic option -S)
        to unset this option set seed to None"""
        if not seed:
            self._seed = None
        else:
            self._seed = str(seed)