#--
#
# Author:
#   Sebastian Tello
#
# Copyright (c) 2003,2012 CORE Security Technologies, CORE SDI Inc.
# All rights reserved.
#
# This computer software is owned by Core SDI Inc. and is
# protected by U.S. copyright laws and other laws and by international
# treaties.  This computer software is furnished by CORE SDI Inc.
# pursuant to a written license agreement and may be used, copied,
# transmitted, and stored only in accordance with the terms of such
# license and with the inclusion of the above copyright notice.  This
# computer software or any other copies thereof may not be provided or
# otherwise made available to any other person.
#
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI Inc. BE LIABLE
# FOR ANY DIRECT,  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR
# CONSEQUENTIAL  DAMAGES RESULTING FROM THE USE OR MISUSE OF
# THIS SOFTWARE
#
#--

import time
import comtypes
import comtypes.client
from comtypes.gen import impact_core_comLib as impact_core_com
from custom_event_handlers import CustomEventHandlerFactory
from impact_manager.impact_lib_exceptions import ImpactModuleException

class ModuleExecutionEventHandler(comtypes.COMObject):
    _com_interfaces_ = [impact_core_com.IModuleExecutionEventHandler]

    def __init__(self, module, general_handler = False):
        self.__status = 'not started'
        self.__task_id = None
        self.__proc_id = None
        self.__event_types = {'output':1,
                              'logline':2,
                              'progress_updated':3,
                              'task_initial':4,
                              'task_running':5,
                              'task_finished':6,
                              'task_aborted':7,
                              'task_stopped':8,
                              'task_stopping':9,
                              'check_udpates_final':10,
                              'task_error':11,
                              'custom':12}

        self.__last_output = None
        self.__last_log = None
        self.__last_progress = None
        self.__is_first_event = True
        self.__module = module
        self.__general_handler = general_handler

    def wrap_me(self):
        return self.QueryInterface(impact_core_com.IModuleExecutionEventHandler)

    def get_status(self):
        return self.__status

    def set_status_failed(self):
        self.__status = "failed"

    def get_task_id(self):
        return self.__task_id

    def set_task_id(self, task_id):
        self.__task_id = task_id

    def get_last_output_time(self):
        return self.__last_output

    def get_last_log_time(self):
        return self.__last_log

    def set_module(self, module):
        self.__module = module

    def get_last_progress_time(self):
        return self.__last_progress

# CUSTOM EVENTS ARE NOW HANDLED IN THE GeneralModuleExecutionEventHandler Class
#    def IModuleExecutionEventHandler_onCustomEvent(self, workspace_id, task_id,
#                                                   event_type, more_in):
#        ## see: http://silo.servers.core-sdi.com/projects/sniper/browser/
#           components/kernel/trunk/interface/impact_client/module_execution.cpp
#
#        if not self.__task_id == task_id:
#            return
#
#        ce_handler = CustomEventHandlerFactory.get_event_handler(event_type,
#                                                                 more_in)
#        a, b = ce_handler.handle()
#        return a


    def IModuleExecutionEventHandler_onEvent(self, event_type, workspace_id,
                                             task_id, more):

        #for general handlers
        if (self.__general_handler and self.__task_id
        and not self.__task_id == task_id):
            return

        if self.__general_handler:
            params = {}
            for mi in more:
                for item in mi:
                    params[item.Key] = item.Value
        else:
            self.__task_id = task_id

        #if this is a general handler, and there's no id set
        #and the event is a task initial whose name matches
        #our module name: Set the id
        if self.__general_handler and not self.__task_id and \
            event_type == self.__event_types['task_initial'] \
            and params['TASK_MODULE'] == self.__module.get_name():
                self.__task_id = task_id

        if not self.__task_id == task_id:
            return

        if event_type == self.__event_types['task_initial']:
            self.__status = "initial"
        elif event_type == self.__event_types['task_finished']:
            self.__status = "finished"
        elif event_type == self.__event_types['task_stopped']:
            self.__status = "stopped"
        elif event_type == self.__event_types['task_aborted']:
            self.__status = "aborted"
        elif event_type == self.__event_types['task_stopping']:
            self.__status = "stopping"
        elif event_type == self.__event_types['task_running']:
            self.__status = "running"
        elif event_type == self.__event_types['output']:
            self.__last_output = time.time()
        elif event_type == self.__event_types['logline']:
            self.__last_log = time.time()
        elif event_type == self.__event_types['progress_updated']:
            self.__last_progress = time.time()

        if self.__is_first_event:
            self.__is_first_event = False
            self.__module.commit_buffered_entity_events_subscriptions()



class UpdateManagerModuleExecutionEventHandler(comtypes.COMObject):
    _com_interfaces_ = [impact_core_com.IModuleExecutionEventHandler]

    def __init__(self):
        self._available_updates = None
        self._ready = False
        self._update_success = False

    def check_updates_status(self, timeout=30):
        start = time.time()
        while not self._ready:
            time.sleep(0.5)
            if (time.time() - start) > timeout and not self._ready:
                raise ImpactModuleException("Timeout reached while "
                                            "checking updates.")
        return self._available_updates

    def get_updates_status(self):
        timeout=1800
        start = time.time()
        while not self._ready:
            time.sleep(0.5)
            if (time.time() - start) > timeout and not self._ready:
                raise ImpactModuleException("Timeout reached while "
                                             "getting updates.")
        return self._update_success

    def wrap_me(self):
        return self.QueryInterface(impact_core_com.IModuleExecutionEventHandler)

    def IModuleExecutionEventHandler_onEvent(self, event_type, workspace_id,
                                                                 task_id, more):
        if event_type in [6,7,8]:
            if event_type == 6:
                self._update_success = True
            self._ready = True

        elif event_type == 10:

            for mi in more:
                for item in mi:
                    if item.Key == "AVAILABLE_UPDATES":
                        self._available_updates = item.Value
            self._ready = True

    def IModuleExecutionEventHandler_onCustomEvent(self, workspace_id, task_id,
                                                 event_type, more_in, more_out):
        ce_handler = CustomEventHandlerFactory.get_event_handler(event_type,
                                                              more_in, more_out)
        a, b = ce_handler.handle()
        return a

